# w.c.s. - web application for online forms
# Copyright (C) 2005-2010  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import os
import socket
import time
import urllib2

try:
    from hashlib import md5 as md5_new
except:
    from md5 import new as md5_new

from quixote.directory import Directory, AccessControlled
from quixote import get_publisher, get_request, get_response
from quixote.html import htmltext, TemplateIO

try:
    import feedparser
except ImportError:
    feedparser = None

from qommon import errors
from qommon.storage import atomic_write

class BackofficeRootDirectory(AccessControlled, Directory):

    @classmethod
    def register_directory(cls, urlname, directory):
        setattr(cls, urlname, directory)

    @classmethod
    def register_menu_item(cls, url, title, check_display_function=None, insert_first=False):
        for i, v in enumerate(cls.menu_items):
            if v[0] == url:
                del cls.menu_items[i]
        if title is None:
            # used to remove an item
            return
        if insert_first:
            cls.menu_items.insert(0, (url, title, check_display_function))
        else:
            cls.menu_items.append((url, title, check_display_function))

    def _q_access(self):
        get_response().breadcrumb.append( ('backoffice/', _('Back Office')) )

        user = get_request().user
        if not user and get_publisher().user_class.count() > 0:
            raise errors.AccessUnauthorizedError(
                    public_msg = _('Access to backoffice is restricted to authorized persons only. '\
                                   'Please login.'))
        if user and not user.can_go_in_backoffice():
            raise errors.AccessForbiddenError()

        get_response().filter['in_backoffice'] = True

    def get_feed_url(self, language):
        feed_url = get_publisher().get_site_option('backoffice_feed_url')
        if feed_url == 'false':
            return None
        elif feed_url:
            return feed_url
        if not get_publisher().backoffice_feed_url:
            return None
        return get_publisher().backoffice_feed_url.get(language)

    def get_feed_filename(self, feed_url):
        md5_hash = md5_new()
        md5_hash.update(feed_url)
        return os.path.join(get_publisher().app_dir, 'feed-' + md5_hash.hexdigest())

    def get_sidebar(self):
        if feedparser is None:
            return None
        feed_url = self.get_feed_url(get_request().language)
        if not feed_url:
            return None
        feed_filename = self.get_feed_filename(feed_url)
        if not os.path.exists(feed_filename):
            self.update_backoffice_feed()
            return None
        if os.stat(feed_filename)[8] < (time.time()-86400):
            self.update_backoffice_feed()

        feed = feedparser.parse(feed_filename)
        if not feed.entries:
            return None

        r = TemplateIO(html=True)
        publication_time = feed.entries[0]['published_parsed']
        if time.mktime(publication_time) > (time.time()-7*86400):
            r += htmltext('<div class="news recent">')
        else:
            r += htmltext('<div class="news">')
        r += htmltext('<h3>%s</h3>') % feed.entries[0]['title'].encode(get_publisher().site_charset)
        r += htmltext(feed.entries[0]['content'][0]['value'].encode(get_publisher().site_charset))
        r += htmltext('</div>')
        return r.getvalue()

    def update_backoffice_feed(self):
        job = get_response().add_after_job(
                    N_('Updating Backoffice Info Feed'),
                    self.update_backoffice_feed_real,
                    fire_and_forget=True)
        job.feed_language = get_request().language

    def update_backoffice_feed_real(self, job):
        feed_url = self.get_feed_url(job.feed_language)
        if not feed_url:
            return None
        try:
            fd = urllib2.urlopen(feed_url)
            s = fd.read()
            fd.close()
        except (urllib2.URLError, socket.error):
            return
        feed_filename = self.get_feed_filename(feed_url)
        atomic_write(feed_filename, s)

    def get_menu_items(self):
        return self.menu_items
