# w.c.s. - web application for online forms
# Copyright (C) 2005-2013  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import datetime
import json
import sys
import traceback
import xml.etree.ElementTree as ET

from quixote.html import TemplateIO, htmltext
from qommon.errors import ConnectionError
from qommon.form import *
from qommon.misc import (http_get_page, http_post_request, get_variadic_url,
        JSONEncoder, json_loads)
from wcs.workflows import (WorkflowStatusItem, register_item_class,
        template_on_formdata, AbortActionException)
from wcs.api import sign_url

TIMEOUT = 30

class JournalWsCallErrorPart: #pylint: disable=C1001
    content = None
    data = None

    def __init__(self, summary, data):
        self.summary = summary
        if data:
            self.data = data[:10000] # beware of huge responses

    def view(self):
        if not get_request().get_path().startswith('/backoffice/'):
            return ''
        r = TemplateIO(html=True)
        r += htmltext('<div class="ws-error">')
        r += htmltext('<h4>%s</h4>') % _('Error during webservice call')
        r += htmltext('<p>%s</p>') % self.summary
        r += htmltext('</div>')
        return r.getvalue()


class WebserviceCallStatusItem(WorkflowStatusItem):
    description = N_('Webservice Call')
    key = 'webservice_call'
    support_substitution_variables = True

    label = None
    url = None
    varname = None
    post = True
    request_signature_key = None
    post_data = None

    action_on_4xx = ':stop'
    action_on_5xx = ':stop'
    action_on_bad_data = ':pass'
    action_on_network_errors = ':stop'
    notify_on_errors = True
    record_errors = False

    def render_as_line(self):
        if self.label:
            return _('Webservice Call "%s"') % self.label
        else:
            return _('Webservice Call')

    def get_parameters(self):
        return ('url', 'post', 'varname', 'request_signature_key', 'post_data',
                'action_on_4xx', 'action_on_5xx', 'action_on_bad_data',
                'action_on_network_errors', 'notify_on_errors',
                'record_errors', 'label')

    def add_parameters_widgets(self, form, parameters, prefix='', formdef=None):
        if 'label' in parameters:
            form.add(StringWidget, '%slabel' % prefix, size=40, title=_('Label'), value=self.label)
        if 'url' in parameters:
            form.add(StringWidget, '%surl' % prefix,
                     title=_('URL'), value=self.url, size=80,
                     hint=_('Common substitution variables are available with the [variable] syntax.'))
        if 'post' in parameters:
            form.add(CheckboxWidget, '%spost' % prefix,
                    title=_('Post formdata (JSON)'),
                    value=self.post)
        if 'post_data' in parameters:
            form.add(WidgetDict, '%spost_data' % prefix,
                    title=_('Post data'),
                    value=self.post_data or {})
        if 'varname' in parameters:
            form.add(VarnameWidget, '%svarname' % prefix,
                     title=_('Variable Name'), value=self.varname)
        if 'request_signature_key' in parameters:
            form.add(StringWidget, '%srequest_signature_key' % prefix,
                    title=_('Request Signature Key'),
                    value=self.request_signature_key)

        error_actions = [(':stop', _('Stop')), (':pass', _('Ignore'))]
        error_actions.extend([(x.id, _('Jump to %s') % x.name) for x in
            self.parent.parent.possible_status])
        for attribute in ('action_on_4xx', 'action_on_5xx', 'action_on_bad_data',
                'action_on_network_errors'):
            if not attribute in parameters:
                continue
            label = {
                    'action_on_4xx': _('Action on HTTP error 4xx'),
                    'action_on_5xx': _('Action on HTTP error 5xx'),
                    'action_on_bad_data': _('Action on non-JSON response'),
                    'action_on_network_errors': _('Action on network errors')
                    }.get(attribute)
            form.add(SingleSelectWidget, '%s%s' % (prefix, attribute),
                    title=label,
                    value=getattr(self, attribute),
                    options=error_actions)

        if 'notify_on_errors' in parameters:
            form.add(CheckboxWidget, '%snotify_on_errors' % prefix,
                    title=_('Notify on errors'),
                    value=self.notify_on_errors)

        if 'record_errors' in parameters:
            form.add(CheckboxWidget, '%srecord_errors' % prefix,
                    title=_('Record errors in the journal'),
                    value=self.record_errors)

    def perform(self, formdata):
        if not self.url:
            # misconfigured action
            return
        url = self.url
        if '[' in url:
            variables = get_publisher().substitutions.get_context_variables()
            url = get_variadic_url(url, variables)

        if self.request_signature_key:
            signature_key = template_on_formdata(formdata, self.request_signature_key)
            if signature_key:
                url = sign_url(url, signature_key)

        headers = {'Content-type': 'application/json',
                'Accept': 'application/json'}
        post_data = None # payload

        # if self.post_data exists, post_data is a dict built from it
        if self.post_data:
            post_data = {}
            for (key, value) in self.post_data.items():
                try:
                    post_data[key] = self.compute(value, raises=True)
                except:
                    get_publisher().notify_of_exception(sys.exc_info())

        # if formdata has to be sent, it's the payload. If post_data exists,
        # it's added in formdata['extra']
        if self.post:
            formdata_dict = formdata.get_json_export_dict()
            if post_data is not None:
                formdata_dict['extra'] = post_data
            post_data = formdata_dict

        try:
            if post_data is not None:
                post_data = json.dumps(post_data, cls=JSONEncoder,
                        encoding=get_publisher().site_charset)
                # increase timeout for huge loads, one second every 65536
                # bytes, to match a country 512kbps DSL line.
                timeout = TIMEOUT
                timeout += len(post_data) / 65536
                response, status, data, authheader = http_post_request(
                        url, post_data, headers=headers, timeout=timeout)
            else:
                response, status, data, auth_header = http_get_page(
                        url, headers=headers, timeout=TIMEOUT)
        except ConnectionError as e:
            status = 0
            self.action_on_error(self.action_on_network_errors, formdata,
                    exc_info=sys.exc_info())

        if self.varname:
            workflow_data = {
                '%s_status' % self.varname: status,
                '%s_time' % self.varname: datetime.datetime.now().isoformat(),
            }
            if status in (204, 205):
                pass # not returning any content
            elif (status // 100) == 2:
                try:
                    d = json_loads(data)
                except (ValueError, TypeError) as e:
                    formdata.update_workflow_data(workflow_data)
                    formdata.store()
                    self.action_on_error(self.action_on_bad_data, formdata,
                            response, data=data, exc_info=sys.exc_info())
                else:
                    workflow_data['%s_response' % self.varname] = d
                    if isinstance(d.get('data'), dict) and d['data'].get('display_id'):
                        formdata.id_display = d.get('data', {}).get('display_id')
                    elif d.get('display_id'):
                        formdata.id_display = d.get('display_id')
            else: # on error, record data if it is JSON
                try:
                    d = json_loads(data)
                except (ValueError, TypeError) as e:
                    pass
                else:
                    workflow_data['%s_error_response' % self.varname] = d
            formdata.update_workflow_data(workflow_data)
            formdata.store()

        if (status // 100) == 4:
            self.action_on_error(self.action_on_4xx, formdata, response, data=data)
        if (status // 100) == 5:
            self.action_on_error(self.action_on_5xx, formdata, response, data=data)

    def action_on_error(self, action, formdata, response=None, data=None, exc_info=None):
        if action in (':pass', ':stop') and (self.notify_on_errors or self.record_errors):
            if exc_info:
                summary = traceback.format_exception_only(exc_info[0], exc_info[1])[-1]
            else:
                summary = ''
                if response:
                    summary = '%s %s' % (response.status, response.reason)
                try:
                    raise Exception(summary)
                except Exception as e:
                    exc_info = sys.exc_info()

            if self.notify_on_errors:
                get_publisher().notify_of_exception(exc_info, context='[WSCALL]')
            if self.record_errors and formdata.evolution:
                formdata.evolution[-1].add_part(JournalWsCallErrorPart(summary, data))
                formdata.store()
        if action == ':pass':
            return
        if action == ':stop':
            raise AbortActionException()
        formdata.status = 'wf-%s' % action
        formdata.store()
        raise AbortActionException()

    def get_target_status(self):
        targets = []
        for attribute in ('action_on_4xx', 'action_on_5xx', 'action_on_bad_data',
                'action_on_network_errors'):
            value = getattr(self, attribute)
            if value in (':pass', ':stop'):
                continue
            targets.append(self.parent.parent.get_status(value))
        return targets

    def get_jump_label(self):
        return _('Error calling webservice')

    def post_data_export_to_xml(self, xml_item, charset, include_id=False):
        if not self.post_data:
            return
        el = ET.SubElement(xml_item, 'post_data')
        for (key, value) in self.post_data.items():
            item = ET.SubElement(el, 'item')
            if type(key) is unicode:
                ET.SubElement(item, 'name').text = key
            elif type(key) is str:
                ET.SubElement(item, 'name').text = unicode(key, charset, 'replace')
            else:
                raise AssertionError('unknown type for key (%r)' % key)
            if type(value) is unicode:
                ET.SubElement(item, 'value').text = value
            elif type(value) is str:
                ET.SubElement(item, 'value').text = unicode(value, charset, 'replace')
            else:
                raise AssertionError('unknown type for value (%r)' % key)


    def post_data_init_with_xml(self, elem, charset, include_id=False):
        if elem is None:
            return
        self.post_data = {}
        for item in elem.findall('item'):
            key = item.find('name').text.encode(charset)
            value = item.find('value').text.encode(charset)
            self.post_data[key] = value

register_item_class(WebserviceCallStatusItem)
